%{
#include <config.h>

#include <stdio.h>
#include <string.h>


static void doparse(void);
int main(int, char**);
static int nextchar(void);
static void recognize(const char*);
int yyparse (void);
static void yywarning (const char*);

#if defined(__cplusplus)
extern "C" {
#endif

static int (*pyylex)(void);
int xxlex(void);
int	zzlex (void);

#if defined(__cplusplus)
}
#endif

int error_count;

/* These variables are defined in biblex.c: */
extern int do_lex_output;
extern long line_number;
extern const char *the_filename;
extern char yytext[BIBYYLMAX];

#define	ERROR_PREFIX	"??"	/* this prefixes all error messages */
#define WARNING_PREFIX	"%%"	/* this prefixes all warning messages */

#define RECOGNIZE(s)	recognize(s)

#define YYDEBUG		1		/* need for -d option support */

#if defined(_CRAY)
extern long yydebug;
#else
extern int yydebug;
#endif

%}

%union
{
  char *sval;
  int ival;
}

%token <sval> TOKEN_ABBREV	1
%token TOKEN_AT		2
%token TOKEN_COMMA	3
%token TOKEN_COMMENT	4
%token <sval> TOKEN_ENTRY	5
%token TOKEN_EQUALS	6
%token <sval> TOKEN_FIELD	7
%token TOKEN_INCLUDE	8
%token TOKEN_INLINE	9
%token <sval> TOKEN_KEY	10
%token TOKEN_LBRACE	11
%token <sval> TOKEN_LITERAL	12
%token TOKEN_NEWLINE	13
%token TOKEN_PREAMBLE	14
%token TOKEN_RBRACE	15
%token TOKEN_SHARP	16
%token TOKEN_SPACE	17
%token TOKEN_STRING	18
%token <sval> TOKEN_VALUE	19

%type <sval> entry_head
%type <sval> assignment_lhs
%type <sval> key_name

 /**********************************************************************
 Without the following precedence specifications, the BibTeX grammar
 is ambiguous, since the two productions

 bibtex_value:		bibtex_value bibtex_space TOKEN_SHARP
				bibtex_simple_value

 bibtex_assignment:	bibtex_assignment_lhs TOKEN_EQUALS bibtex_value

 result in a shift/reduce conflict when the input stack contains

	bibtex_assignment_lhs TOKEN_EQUALS bibtex_value

 and the next token is one of the bibtex_space tokens.  We could shift
 that token to match the start of the bibtex_value production, or we
 could reduce to bibtex_assignment using the second production.

 The way to remove the ambiguity is described in the book

 @String{pub-PH          = "Pren{\-}tice-Hall"}
 @String{pub-PH:adr      = "Englewood Cliffs, NJ 07632, USA"}

 @Book{Holub:CDC90,
  author =       "Allen I. Holub",
  title =        "Compiler Design in {C}",
  publisher =    pub-PH,
  address =      pub-PH:adr,
  year =         "1990",
  note =         pub-PH # " Software Series, Editor: Brian W.
                 Kernighan.",
  ISBN =         "0-13-155045-4",
 }

 on pp. 871--875.
 **********************************************************************/

%nonassoc TOKEN_EQUALS
%left TOKEN_SPACE TOKEN_INLINE TOKEN_NEWLINE
%left TOKEN_SHARP

%%
file:		  opt_space
			{RECOGNIZE("file-1");}
		| opt_space object_list opt_space
			{RECOGNIZE("file-2");}
		;

object_list:	  object
			{RECOGNIZE("object-1");}
		| object_list opt_space object
			{RECOGNIZE("object-2");}
		;

object:	  	  TOKEN_AT opt_space at_object
			{
			    RECOGNIZE("object");
			}
		;

at_object:	  comment
			{  
			  RECOGNIZE("comment");
			}
		| entry
			{
			  RECOGNIZE("entry");
			}
		| include
			{
			  RECOGNIZE("include");
			}
		| preamble
			{
			  RECOGNIZE("preamble");
			}
		| string
			{
			  RECOGNIZE("string");
			}
		| error TOKEN_RBRACE
			{
			  RECOGNIZE("error");
			}
		;

comment:	  TOKEN_COMMENT 
                        {
			  printf("<Comment>");
			}
                        opt_space
			TOKEN_LITERAL
			{
			  printf("%s</Comment>",$4);
			  RECOGNIZE("comment");
			}
		;

entry:		  entry_head
			assignment_list
			TOKEN_RBRACE
			{
			  RECOGNIZE("entry-1");
			}
		| entry_head
			assignment_list
			TOKEN_COMMA opt_space
			TOKEN_RBRACE
			{
			  RECOGNIZE("entry-2");
			  printf("</%s>\n\n",$1);
			}
		| entry_head TOKEN_RBRACE
			{
			  RECOGNIZE("entry-3");
			}
		;

entry_head:	  TOKEN_ENTRY
                        {
			    printf("<%s>\n",$1);
			    $$=strdup($1);
			}
                        opt_space
			TOKEN_LBRACE opt_space
			key_name
			{
			    printf("<label>%s</label>\n",$6);
			}
                        opt_space
			TOKEN_COMMA opt_space
			{
			    RECOGNIZE("entry_head");
			}
		;

key_name:	  TOKEN_KEY
			{
			  $$=strdup($1);
			  RECOGNIZE("key_name-1");
			}
		| TOKEN_ABBREV
			{
			  RECOGNIZE("key_name-2");
			}
		;

include:	  TOKEN_INCLUDE
                        {
			  printf("<Include>");
			}
                        opt_space
			TOKEN_LITERAL
			{
			  printf("<Include>");
			  RECOGNIZE("include");
			}
		;

preamble:	  TOKEN_PREAMBLE
                        {
			  printf("<Preamble>");
			}
                        opt_space
			TOKEN_LBRACE opt_space
			value opt_space
			TOKEN_RBRACE
			{
			  printf("</Preamble>");
			  RECOGNIZE("preamble");
			}
		;

string:		  TOKEN_STRING
                        {
			  printf("<String>");
			}
                        opt_space
			TOKEN_LBRACE opt_space
			assignment
			opt_space TOKEN_RBRACE
			{
			  printf("</macrotext></String>\n\n");
			  RECOGNIZE("string");
			}
		;

value:	  	  simple_value
			{
			    RECOGNIZE("value-1");
			}
		| value opt_space
			{RECOGNIZE("value-1-1");}
			TOKEN_SHARP
			{RECOGNIZE("value-1-2");}
			opt_space simple_value
			{RECOGNIZE("value-2");}
		;

simple_value:	  TOKEN_VALUE
			{
			    printf("%s",$1);
			    RECOGNIZE("simple_value-1");}
		| TOKEN_ABBREV
			{
			    printf("<ref>%s</ref>",$1);
			    RECOGNIZE("simple_value-2");
			}
		;

assignment_list:  assignment
			{RECOGNIZE("single assignment");}
		| assignment_list
			TOKEN_COMMA opt_space
			assignment
			{RECOGNIZE("assignment-list");}
		;

assignment:	  assignment_lhs opt_space
			TOKEN_EQUALS opt_space
			{RECOGNIZE("assignment-0");}
			value opt_space
			{
			    if(strcmp($1,"donotclose"))/*caccamo*/
				printf("</%s>\n",$1);
			    RECOGNIZE("assignment");
			}
		;

assignment_lhs:	  TOKEN_FIELD
			{
                           printf("<%s>",$1);
			   $$=strdup($1);
                           RECOGNIZE("assignment_lhs-1");
                         }
		| TOKEN_ABBREV
			{
			  if(!strcmp($1,"comment"))
			    printf("<comment>%s</comment>",$1); /*caccamo*/
			  else
			    printf("<alias>%s</alias><macrotext>", $1);
			  $$="donotclose";/*caccamo*/
			  RECOGNIZE("assignment_lhs-2");
			}
		;

opt_space:	/* empty */
			{RECOGNIZE("opt_space-1");}
		| space
			{RECOGNIZE("opt_space-2");}
		;

space:		  single_space
			{RECOGNIZE("single space");}
		| space single_space
			{RECOGNIZE("multiple spaces");}
		;

single_space:	  TOKEN_SPACE
		| TOKEN_INLINE
		| TOKEN_NEWLINE
		;
%%

/* end of grammar */

static void doparse(void)
{
    /* parse a complete BibTeX file or BibTeX token stream */
    int c;

    line_number = 1L;

    /* A single character lookahead distinguishes between a token
    stream, which starts with a line number directive, and a BibTeX
    file, which does not.  We then pick a suitable lexical analyzer
    function. */

    c = getchar();
    (void)ungetc(c, stdin);
    
    pyylex = (c == '#') ? zzlex : xxlex;
    

    yyparse();				/* this handles the entire file */
}


int main(int argc, char *argv[])
{
    int code; /* error return code */
    int k;    /* index into argv[] */
    int n;    /* index into argv[] */
    FILE *fp; /* pointer to current input file */

    do_lex_output = 0;			/* prevent output from xxlex() */
    error_count = 0;
    for(k = 1; k < argc; ++k) {					/* check for command-line options */
        if((strcmp(argv[k],"-d") == 0) || (strcmp(argv[k],"-D") == 0)) {
            yydebug = 1;
            do_lex_output = 1;
            for (n = k; n < argc; ++n)	/* shuffle remaining arguments down */
                argv[n] = argv[n+1];
            argc--;			/* and reduce argument count */
        }
    }
    printf("<?xml version=\"1.0\" encoding=\"UTF-8\"?><bibxml>");
    if (argc > 1) {		/* files to parse named on command line */
        for (k = 1; k < argc; ++k) {
            fp = freopen(argv[k], "r", stdin);
            if (fp == (FILE*)NULL) {
                (void)fprintf(stderr,
                              "\n%s Ignoring open failure on file [%s]\n",
                              ERROR_PREFIX, argv[k]);
                (void)perror("perror() says");
            } else {
                the_filename = argv[k];
                doparse();
                (void)fclose(fp);
            }
        }
    } else {				/* parse stdin */
        the_filename = "stdin";
        doparse();
    }
    printf("</bibxml>\n");
    return (error_count > 0) ? EXIT_FAILURE : EXIT_SUCCESS;
}


static int nextchar(void)
{
    int c;

    c = getchar();
    if (yydebug)
        putchar (c);
    return c;
}


static void recognize(const char *s)
{
    if (yydebug)
        printf("[%s]\n", s);
}


void yyerror(const char *s)
{
    error_count++;
    (void)fflush(stdout);
    (void)fprintf(stderr,"%s \"%s\", line %ld: %s\tNext token = \"%s\"\n",
                  ERROR_PREFIX, the_filename, line_number, s, yytext);
    (void)fflush(stderr);
}


int yylex(void)
{
    int v;
    v = (*pyylex) ();

    /* xxlex() (== yylex() in biblex.c) biases token values by 1000, but
       the token streams produced by biblex and "bibclean --no-prettyprint"
       exclude that bias, for compact output. */

    return ((pyylex == zzlex) ? v : (v - 1000));
}


static void yywarning(const char *s)
{
    (void)fflush(stdout);
    (void)fprintf(stderr,"%s %s\tNext token = \"%s\"\n",
		  WARNING_PREFIX, s, yytext);
    (void)fflush(stderr);
}


int zzlex(void)
{
    char buffer[BUFSIZ];	/* only needs space for a single number token */
    char *limit;		/* how far we can store into buffer[] */
    char *p;			/* pointer into buffer[] */
    int c;			/* current input character */

    limit = &buffer[sizeof(buffer) - 1]; /* address of last slot in buffer[] */

    for (;;) {			/* `infinite' loop for ignoring comments and empty lines */
        while (((c = nextchar ()) > 0) && isspace (c))
            ;			/* skip leading whitespace */

        /* Collect first non-blank token on line, truncating if too big for buffer[] */
        p = &buffer[0];
        *p++ = c;
        while (((c = nextchar ()) > 0) && !isspace (c)) {
            if (p < limit)
                *p++ = c;
        }
        *p = '\0';

        while ((c != '\n') && ((c = nextchar ()) > 0))
            ;			/* skip rest of line */

        if (c == '\n')		/* should always be true */
            line_number++;

        if (buffer[0] == '#')	/* comment: pseudo return tail-recursively */
            continue;
        else if (isdigit ((int) buffer[0]))	/* token number */
            return ((int) strtol ((const char *) &buffer[0], &limit, 10));
        else if (c == '\n')	/* empty line: pseudo return tail-recursively */
            continue;
        else if (feof (stdin))	/* assume end of file */
            return (0);		/* standard lex end-of-file signal */
        else
            return (-1);	/* should never happen */
    }
}
